/*
** $Log:   C:/shell/vcs/dirfuncs.c_v  $
**
**   Rev 3.7   02 Oct 1986 13:13:32   kent
**lint pass over everything
**
**   Rev 3.6   01 Oct 1986 14:29:12   kent
**Fixed bug in pushd with no args
**
**   Rev 3.5   01 Oct 1986 13:31:10   kent
**Put all globals into shell.h
**
**   Rev 3.4   01 Oct 1986  9:25:06   kent
**Initial revision.
*/
#include "shell.h"
#define CHDIR	0x3b
#define RMDIR	0x3a
#define MKDIR	0x39
void print_error();
int cd(argc,argv)
char *argv[];
{
	static char *usage = "usage : cd newdir\r\n";
	if (argc == 1)
	{
		write(2,usage,strlen(usage));
		return 0;
	}
	return dirfunc("chdir",*(++argv),CHDIR);
}

int md(argc,argv)
	char *argv[];
{
	static char *usage = "usage : md newdir\r\n";

	if (argc == 1)
	{
		write(2,usage,strlen(usage));
		return 0;
	}
	return dirfunc("mkdir",*(++argv),MKDIR);
}

int rd(argc,argv)
char *argv[];
{
	static char *usage = "usage : rmdir directory";
	if (argc == 1)
	{
		write(2,usage,strlen(usage));
		return 0;
	}
	return dirfunc("rmdir",*(++argv),RMDIR);
}

int dirfunc(name,string,function)
char *name, *string;
{
	struct registers regs;
#ifdef AZTEC86
	extern unsigned _dsval;
	regs.ds = (int)_dsval;
#define CARRY 1
#endif
#ifdef MSC
	GETSEGS(&regs);
#endif
	regs.ax = (int)((unsigned)function << 8);
	regs.dx = (int)string;
#ifdef AZTEC86
	if(sysint21(&regs,&regs) & CARRY)
#endif
#ifdef MSC
	sysint21(&regs,&regs);
	if (regs.cflag)
#endif
	{
		print_error(name,regs.ax);
		return -1;
	}
	return 0;
}

static char *dirstack[10];
static int dsp = -1;

int pushd(argc,argv)
char *argv[];
{
	char *getcwd(), *savestr();
	char *pusherr = "pushd : dir stack overflow";
	static char dirbuf[64];
#ifdef AZTEC86
#define BUFSTART 1
	dirbuf[0] = '\\';
#endif
#ifdef MSC
#define BUFSTART 0
#endif
	if (NULL ==  getcwd(&dirbuf[BUFSTART],64))
	{
		perror("pushd");
		return -1;
	}
	if (argc == 1)
	{
		if (-1 == dsp)
		{
			fprintf(stderr,"pushd : dir stack underflow\n");
			return -1;
		}
		if (-1 == dirfunc("pushd",dirstack[dsp],CHDIR))
			return -1;
		dirstack[dsp] = savestr(dirbuf);
		return 0;
	}
	if (++dsp == 10)
	{
		write(2,pusherr,strlen(pusherr));
		crlf();
		--dsp;
		return -1;
	}
	if (-1 == dirfunc("pushd",*(++argv),CHDIR))
	{
		--dsp;
		return -1;
	}
	dirstack[dsp] = savestr(dirbuf);
	return 0;
}

int popd()
{
	register int returnval = 0;
	static char *poperr = "popd : dir stack underflow";
	if (dsp == -1)
	{
		write(2,poperr,strlen(poperr));
		crlf();
		return -1;
	}
	if (-1 == dirfunc("popd",dirstack[dsp],CHDIR))
	{
		returnval = -1;
	}
	free(dirstack[dsp--]);
	return returnval;
}

int dirs()
{
	int sp = dsp;
	while (sp >= 0)
	{
		write(1,dirstack[sp],strlen(dirstack[sp]));
		crlf();
		--sp;
	}
	return 0;
}

int pwd()
{
	char *getcwd();
	register char *dir;
	char *dirfix;
	if (!(dir = getcwd(NULL,64)))
	{
		perror("pwd");
		return -1;
	}
	/* change backslashes to forward slashes */
	for (dirfix = dir; *dirfix; dirfix++)
		*dirfix = (*dirfix == '\\' ? '/' : tolower(*dirfix));
	write(1,"/",1);
	write(1,dir,strlen(dir));
	crlf();
	if (dir)
		free(dir);
	return 0;
}
