/*
** $Log:   C:/shell/vcs/expand.c_v  $
**
**   Rev 3.6   02 Oct 1986 13:12:38   kent
**lint pass over everything
**
**   Rev 3.5   01 Oct 1986 13:30:36   kent
**Put all globals into shell.h
**
**   Rev 3.4   01 Oct 1986  9:24:40   kent
**Initial revision.
** expand.c - three substitution functions for the shell
** aexpand - expand alii from the environment
** vexpand - expand variables from the environment
** hexpand - expand history references
** fexpand - expand ambiguous file names
*/
/* #define DEBUG */
#include "shell.h"

typedef unsigned u;

/*
** local versions of library functions
*/
char *get_env();
#define exit(code)	my_exit(code)
#define getenv(s)	get_env(s)


/* dump_current - debug function to dump out current history environment */
#ifdef DEBUG
void dump_current(s)	
	char *s;
{
	register int i;
	fprintf(stderr,"buffer=%s\n",s);
	fprintf(stderr,"_current->j = %d\n",_current->j);
	fprintf(stderr,"_current->hiscount = %d\n",_current->hiscount);
	fprintf(stderr,"_current->histsize = %d\n",_current->histsize);
	for(i=0; i < _current->histsize; i++)
	fprintf(stderr,"_current->history[%d] = %s\n",i,_current->history[i]);
}
#endif

/*
** addtoline
** concatenates an argument onto the end of a string, dynamically expanding
** the string as need be.
*/
char *addtoline(newline,cur_arg,add_space)
	char *newline; char *cur_arg; int add_space;
{
	char *savestr();

	if (NULL == newline)
		newline = savestr(cur_arg);
	else
	{
		newline =
			MustRealloc(newline,(u)(strlen(newline)+strlen(cur_arg)+2));
		if (add_space) strcat(newline," ");
		strcat(newline,cur_arg);
	}
	return newline;
}

/*
** redirect - do all i/o redirection
*/

char here_dirty = 0;
int in = -1, out = -1;
void redirect_cleanup()
{
	if (here_dirty)
		unlink(herename);
	here_dirty = 0;
	if (in != -1)
		close(in);
	if (out != -1)
		close(out);
}

char *
redirect (buf)
char *buf;
{
	char *nc,*newline = NULL;
	register char *cur_arg;
	register unsigned int i;
	char noclobber = 0;
	int k,omode; char *fname;
	char err_also = 0;

	in = out = -1;
	/* set noclobber switch from environment */
	if (NULL != (nc = get_env("NOCLOBBER")))
		noclobber = *nc == '1';
	/* look through all the args in the buffer */
	for (i = 0;;i++)
	{
		if (NULL == (cur_arg = ntharg(buf,i)))
			break;
		if (BADQUOTES == cur_arg)
		{
			fprintf(stderr,"No closing quotes on command line\n");
			if (newline) free(newline);
			return NULL;
		}
		switch(*cur_arg)
		{
		case '>':
			/* you can redirect output '>',
			    append '>>',
				redirect stderr too '>&',
				redirect stderr and append '>>&'
			 */
			k = 1;
			++cur_arg;
			switch(*cur_arg)
			{
			case '>':
				++cur_arg;
				omode = O_CREAT|O_WRONLY|O_APPEND|O_TEXT; 
				if (cur_arg[1] == '&')
				{
					err_also++;
					++cur_arg;
				}
				if (cur_arg[1] == '!')
				{
					noclobber = 0;
					++cur_arg;
				}
				break;
			case '&':
				++cur_arg;
				err_also++;
			default:
				omode = O_CREAT|O_WRONLY|O_TRUNC|O_TEXT;
				if (cur_arg[1] == '!')
				{
					noclobber = 0;
					++cur_arg;
				}
			}
			goto redirect;
		case '<':
			++cur_arg;
			if (*cur_arg == '<')
			{
			/* here document code.  inputfile is set in
			** stdsave before it reopens stdin with the console
			** we use that handle to read the script file
			** using getline from main.s
			*/
				extern int inputfile;
				FILE *op,*fopen(),*fdopen();
				char *buf,*fgets();
				int result;
				++cur_arg;
				buf = MustMalloc((u)128);
				if (*cur_arg)
					fname = cur_arg;
				else
				{
					++i;
					fname = ntharg(buf,i);
				}
				if (NULL == (op = fopen(herename,"w")))
				{
					Free(buf);
					fprintf(stderr,"can't open %s\n",herename);
					goto err_out;
				}
				while (0 != (result = getline(inputfile,buf)))
				{
					if (0==strncmp(fname,buf,strlen(fname)))
						break;
					strcat(buf,"\n"); /* add a line feed */
					fputs(buf,op);
				}
				here_dirty++;
				fclose(op);
				Free(buf);
				if (result == 0)
				{
					fprintf(stderr,"no termination of here document\n");
					goto err_out;
				}
				close(0);
				in = k = open(herename,O_RDONLY);
				if (k == -1)
				{
					perror(herename);
					goto err_out;
				}
				continue;
			}
			k = 0;
	redirect:
			if (*cur_arg)
				fname = cur_arg;
			else
			{
				char *quotechar;
				i++;
				fname = ntharg(buf,i);
				if ((quotechar = strchr("\'\"",fname[0])) != NULL)
				{
					++fname;
					if (NULL == (quotechar = strchr(fname,*quotechar)))
					{
						fprintf(stderr,"bad redirect name %s\n",fname-1);
						goto err_out;
					}
					*quotechar = '\0';
				}
			}
			close(k);
			if (k)
			{
				if (0 != (omode & O_TRUNC) && noclobber && filep(fname))
				{
					fprintf(stderr,"redirect : %s exists\n",fname);
					goto err_out;
				}
				if (0 != (omode & O_APPEND) && noclobber && !filep(fname))
				{
					fprintf(stderr,"redirect : %s doesn't exist\n",fname);
					goto err_out;
				}
				out = k = open(fname,omode);
			}
			else
				in = k = open(fname,O_RDONLY);
			if (k == -1)
			{
				perror(fname);
err_out:
				if (newline)
					Free(newline);
				return NULL;
			}

			if (err_also)
			{
				close(2);
				(void)dup(1);
			}
			/* go back for next argument */
			continue;
		default:	/* add the current argument to the command line */
			newline = addtoline(newline,cur_arg,TRUE);
		}
	}
	return newline;
}
/*
** aexpand - expand all alias references
*/
char *
aexpand(buf)
char *buf;
{
	char *newline = NULL;
	register char *cur_arg;
	register int i;
	char *envstring;
#ifdef STANDALONE
	_current->history[_current->j] = buf;
#endif
	/* look through all the args in the buffer */
	for (i = 0;;i++)
	{
		if (NULL == (cur_arg = ntharg(buf,(unsigned)i)))
			break;
		if (BADQUOTES == cur_arg)
		{
			fprintf(stderr,"No closing quotes on command line\n");
			if (newline) free(newline);
			return NULL;
		}
		if (i == 0 && NULL == strchr("\"'",*cur_arg))
		{
			char *alias = MustMalloc((unsigned)strlen(cur_arg)+2);
			sprintf(alias,"`%s",cur_arg);

			if (NULL != (envstring = getenv(alias)))
				newline = addtoline(newline,envstring,TRUE);
			else
				newline = addtoline(newline,cur_arg,TRUE);
			Free(alias);
		}
		else
		{
			/* otherwise, add the arg itself */
			newline = addtoline(newline,cur_arg,TRUE);
		}
	}
	return newline;
}

char *envname(str)
	register char *str;
{
	register char *lp;
	static char local[64];
	lp = local;
	if (*str=='$')
	{
		strcpy(local,"$");
		return local;
	}
	while (*str)
	{
		if (!isalnum(*str))
		{
			*lp = '\0';
			return local;
		}
		*lp++ = *str++;
	}
	*lp = '\0';
	return local;
}

char *string(c)
	char c;
{
	static char foo[2];
	foo[0] = c; foo[1] = '\0';
	return foo;
}

/*
** vexpand - expand all variable references
*/
char *
vexpand(buf)
char *buf;
{
	char *newline = NULL;
	register char *cur_arg;
	register int i;
	char *envstring;
	char *strchr();
#ifdef STANDALONE
	_current->history[_current->j] = buf;
#endif
	/* look through all the args in the buffer */
	for (i = 0;;i++)
	{
		char *current; char *cp;
		if (NULL == (cur_arg = ntharg(buf,i)))
			break;
		if (BADQUOTES == cur_arg)
		{
			fprintf(stderr,"No closing quotes on command line\n");
			if (newline) free(newline);
			return NULL;
		}
		current = savestr(cur_arg);	/* save current argument */
		cp = current;
		if (i != 0)
			newline = addtoline(newline," ",FALSE);
		if (*cp != '\'')
		{
			char inquotes = *cp == '"';
			while (*cp)
			{
				switch (*cp)
				{
				case '\\':
					if (!inquotes)
					{
						newline = addtoline(newline,string(*cp++),FALSE);
						newline = addtoline(newline,string(*cp++),FALSE);
					}
					else
					{
						char addchar = *++cp;
						switch(*cp++)
						{
						case 'b':
							addchar = '\b';
							break;
						case 'r':
							addchar = '\r';
							break;
						case 'n':
							addchar = '\n';
							break;
						case 'f':
							addchar = '\f';
							break;
						case 'a':	/* alert character */
							addchar = '\07';
							break;
						}
						newline = addtoline(newline,string(addchar),FALSE);
					}
					break;
				case '$':
					++cp;
					if (NULL != (envstring = get_env(envname(cp))))
						newline = addtoline(newline,envstring,FALSE);
					else if (strcmp(envname(cp),"$") == 0)
						newline = addtoline(newline,"$",FALSE);
					cp += strlen(envname(cp));
					break;
				default:
					newline = addtoline(newline,string(*cp),FALSE);
					cp++;
					break;
				}
			}
		}
		else
			newline = addtoline(newline,cp,FALSE);
		free(current);
	}
	return newline;
}

char *
fexpand(buf)
char *buf;
{
	char *newline = NULL;
	register char *cur_arg;char *wildmatch;
	register int i;
#ifdef STANDALONE
	_current->history[_current->j] = buf;
#endif
	/* look through all the args in the buffer */
	for (i = 0;;i++)
	{
		if (NULL == (cur_arg = ntharg(buf,i)))
			break;
		if (BADQUOTES == cur_arg)
		{
			fprintf(stderr,"No closing quotes on command line\n");
			if (newline) free(newline);
			return NULL;
		}
		/* if we have a wildcard */
		if (
			(NULL != strchr(cur_arg,'*') || NULL != strchr(cur_arg,'?')) &&
			NULL == strchr("\"'",cur_arg[0])
		)
		{
			char *path;
			/* get a copy of the path component */
			path = path_part(cur_arg);
			/* while there are matches, add them to the line */
			if (NULL != (wildmatch = get_first_name(cur_arg)))
			{
				do
				{
					if (path)	/* if there is a path */
					{	/* add it in first */
						newline = addtoline(newline,path,TRUE);
						newline = addtoline(newline,wildmatch,FALSE);
					}
					else
						newline = addtoline(newline,wildmatch,TRUE);
				} while (NULL != (wildmatch = get_next_name()));
			}
		}
		else
		{
			/* otherwise, add the arg itself */
			newline = addtoline(newline,cur_arg,TRUE);
		}
	}
	return newline;
}

char *
hexpand(buf)
char *buf;
{
	char *newline = NULL;
	register char *cur_arg,*argptr;
	register int i;
	int histindex,argindex,repeat,takeline;
#ifdef STANDALONE
	/* just for debugging */
	_current->history[_current->j] = buf;
#endif
	/* look through all the args in the buffer */
	for (i = 0;;i++)
	{
		if (NULL == (cur_arg = ntharg(buf,i)))
			break;
		if (BADQUOTES == cur_arg)
		{
			fprintf(stderr,"No closing quotes on command line\n");
			if (newline) free(newline);
			return NULL;
		}
		if (*cur_arg == '!')
		{
			char *foo = MustMalloc((unsigned)strlen(cur_arg)+1);
			strcpy(foo,cur_arg);
			cur_arg = foo;
			takeline = 0;
			++cur_arg;	/* bump past bang */
			/* parse history expression */
			switch (*cur_arg)
			{
			case '!':	/* last command line */
			set_histindex:
				if (_current->j)	/* special case first time through */
				{
					histindex =  (int)(_current->hiscount ? 
					_current->hiscount - 1 :  _current->histsize - 1);
				}
				else
				{
					goto error_out;
				}
				++cur_arg;	/* bump past bang */
				break;
			case '#':	/* look through current command line */
				histindex = (int)_current->hiscount; /* use current command */
				++cur_arg;	/* bump past sharp */
				break;
			case '$':	/* !$ synonym for !!:$ */
			case '^':	/* !^ synonym for !!:^ */
			case '*':	/* !* synonym for !!:* */
				--cur_arg;	/* back up one to simulate !! */
				goto set_histindex;	/* branch back to set up histindex */
			default:
				if (isdigit(*cur_arg)||*cur_arg=='-')
				{
					/* repeat numbered command */
					repeat = atoi(cur_arg);
					if (repeat < 0)	/* handle relative addressing */
						repeat += (int)_current->j;

					/* if command is within range */
					if ((_current->j - repeat) < _current->histsize 
						&& repeat < _current->j)
					{
						histindex = (int)(repeat % _current->histsize);
					}
					else
					{
						goto error_out;	
					}
					/* skip the number */
					if(*cur_arg == '-') ++cur_arg;	/* leading sign */
					while (isdigit(*cur_arg))
						++cur_arg;
				}
				else
				{
					/* look for previous command starting with the string
					** after the bang
					*/
					unsigned int j; int k;
					unsigned lowlimit;
					j = _current->j % _current->histsize; /* index in array */
					if (_current->j < _current->histsize)
						/* we only search from 0 to j */
						lowlimit = 0;
					else
					{
						/* if we are on last entry in history list */
						if (j == (_current->histsize - 1))
							lowlimit = 0;
						else
							lowlimit = j + 1;
					}
					for (k = lowlimit; k != j ;k=(k+1) % _current->histsize)
							if (0 == strncmp(_current->history[k],
								cur_arg,strlen(cur_arg)))

					{
								histindex = k;
								break;
					}
					if (k == j)	/* didn't find it */
					{
						goto error_out;
					}
					/* go past end of string */
					while (*cur_arg)
						++cur_arg;
				}
			}
			/* look for particular argument substitutions */
			switch (*cur_arg)
			{
			case '$':
			case '*':
			case '^':
				goto nocolon;
			case ':':
				++cur_arg;	/* point past colon */
		nocolon:
				switch (*cur_arg)
				{
				case '^':
					++cur_arg;
					argindex = 1;
					break;
				case '$':
					if (0 > (argindex = lastarg(_current->history[histindex])))
					{
						goto error_out;
					}
					++cur_arg;
					break;
				case '*':
					takeline = 1;	/* take arg 1 through arg n */
					argindex = 1;
					++cur_arg;
					break;
				default:
					if (isdigit(*cur_arg))
					{
						/* index of argument */ 
						argindex = atoi(cur_arg);
						while(isdigit(*cur_arg))
							++cur_arg;
						if (*cur_arg == '*')
						{
							++cur_arg;
							takeline = 1;
						}
					} 
					else
					{	
	error_out:
						fprintf(stderr,"bad history expression\n");
						Free(newline);  free(foo);
						return NULL;
					}
				}
				break;
			default:
				takeline = 1;
				argindex = 0;
			}
			/* pick up pointer to argument in history we need */
			if (takeline == 0)
			{
				if (NULL == 
				(argptr = ntharg(_current->history[histindex],argindex)))
					goto fooFree;
				newline = addtoline(newline,argptr,TRUE);
			}
			else
			{
				while (NULL !=
				(argptr = ntharg(_current->history[histindex],(u)argindex++)))
					newline = addtoline(newline,argptr,TRUE);
			}
			/* pick up trailing argument component */
			if (!isspace(*cur_arg) && '\0' != *cur_arg)
				newline = addtoline(newline,cur_arg,FALSE);
fooFree:
			Free(foo);	/* get rid of copy of current argument */
		}
		else
			/* otherwise, add the arg itself */
			newline = addtoline(newline,cur_arg,TRUE);
	}
	return newline;
}

