.TH EXPR 1 "Expression Evaluator"
.SH EXPR
expr - expression evaluator
.SH SYNOPSIS
.bo
expr
[-[hovr]] expression
.sp 1
.in +5
.nf
-h print result in hexidecimal
-o print result in octal
-d print result in decimal (default)
-r return the result as the result code to DOS
.br
.in -5
.fi
If no expression is given on the command line, then expr will go into
a mute, command driven mode - you type an expression followed by a
carriage return, and it prints out the answer.
.sp 1
If you specify an expression on the command line that has characters
in it that DOS interprets to mean redirection or file name expansion,
enclose them in quotes, single or double.
.sp 1
.SH DESCRIPTION
expr is a small expression evaluator, suitable for evaluating
expressions consisting of string and integer operands.  It uses a
simple, fairly stupid recursive descent parser to evaluate the expression.
.sp 1
.nf
.in +5
.nf
The grammar used by expr is as follows
.sp 1
EXPR ::= TERM  | TERM BINOP EXPR
TERM ::= ATOM | EXPR | '(' EXPR ')' | '!' TERM | '-' TERM
ATOM ::= INTEGER | STRING
INTEGER ::= [0-9]* | 0[xX][0-9a-fA-F]* | 0[0-7]*
STRING ::= 	[^0-9=!<>&|-+/%*()][^=!<>&|-+/%*()]*
BINOP ::=  '==' | '=' | '!=' | '>=' | '<=' | 
.in +10
'<>' | '&&' | '||' | '<' | '>' | '+' | 
'-' | '*' | '%' | '&' | '|'
.sp 1
.br
.in -15
.fi
There is no operator precedence, except that enforced by parentheses.
Mathematical operations on string operands are errors, and terminate
.bo
expr
with a result code of -1.
.sp
The result of legal binary operations on string operands are typed as
integers, so you can do silly things like "expr (HELLO==HELLO)+2." if
you wish.
.sp
.SH CAVEATS
As
.bo
expr
has no notion of precedence, when in doubt, parenthesize. 
.SH EXAMPLES
.nf
.in +5
C>expr HELLO==GOODBYE
0
C>expr "0x10*1024"
16384
C>expr -h "1024 * 640"
a0000
.sp 1
.in -5
.SH FILES
an.nro
.SH "RETURN VALUE"
Returns -1 if an error occurs, 0 if 
.bo
-r
isn't specified, or the return value of the command line expression,
if -r is specified.
.SH BUGS
As mentioned above this expression evaluator evaluates strictly left
to right, and has no notion of precedence.  This means that you must
parenthesize nearly all complex expressions.  I will leave it as an
exercise to the reader to introduce precedence.
.SH MISCELLANEOUS
If you comment out function
.bo
main
you can call expr directly.  It returns a long integer indicating the
result of the expression.
