/*
** $Log:   C:/shell/vcs/ls.c_v  $
**
**   Rev 3.6   02 Oct 1986 13:10:04   kent
**lint pass over everything
**
**   Rev 3.5   01 Oct 1986 13:28:38   kent
**Put all globals into shell.h
**
**
**   Rev 3.4   01 Oct 1986  9:23:48   kent
**Initial revision.
*/
#ifdef MAIN
#undef MAIN
#define MAIN_DEFINED
#endif
#include "shell.h"
#ifdef MAIN_DEFINED
#define MAIN
#endif

#define ISROOT(s) (0==strcmp(s,"/"))

typedef struct
{
	char attribute;
	unsigned file_time;
	unsigned file_date;
	long file_size;
	char file_name[13];
} file_desc;

typedef struct
{
	char dos_reserved[21];
	char attribute;
	unsigned file_time;
	unsigned file_date;
	long file_size;
	char file_name[13];
} fcb;

#define maxfiles 128 

file_desc *getfirst(),*getnext();
char *strchr(), *strrchr();

static int mode = 0x10;
static int lsverbose=0,column=4,recurse=0;
static int lsquiet = 0;
static int drivenum = 0;

/* sort types */
#define ALPHA 0
#define RALPHA 1
#define TIME 2
#define RTIME 3
int sortype;
/* sort functions */
int files_cmp(),times_cmp();
int rfiles_cmp(),rtimes_cmp();


/* array of pointers to sort functions - used for calling qsort */
int (*compare[])() = { files_cmp,rfiles_cmp,times_cmp,rtimes_cmp };

#ifdef AZTEC86
/* paramaterized printf - use printf if i/o is redirected, scr_printf otherwise */
static char buffer[128];
static char *charptr;
static void x(c)
{
#ifndef ASM
	if (c=='\n')
		*charptr++ = '\r';
	*charptr++ = c;
#else
#asm
	mov	al,byte ptr 4[bp]
	mov	bx,word ptr charptr_
	cmp	al,10
	jne	onechar
	mov	byte ptr [bx],13
	inc	bx
onechar:
	mov	[bx],al
	inc	bx
	mov charptr_,bx
#endasm
#endif
}
int printf(fmt,args)
	char *fmt; unsigned args;
{
	charptr = buffer;
	format(x,fmt,&args);
	*charptr = '\0';
	write(1,buffer,(int)(charptr-buffer));
}
#ifndef MAIN
int fprintf(f,fmt,args)
FILE *f; char *fmt; unsigned args;
{
	fflush(f);	/* clear any pending i/o so fprintf won't hose us */
	charptr = buffer;
	format(x,fmt,&args);
	*charptr = '\0';
	write(fileno(f),buffer,(int)(charptr-buffer));
}
#endif
#endif

#ifdef MSC
extern int printf();
#endif
int (*pf)() = printf;

long time();
short year;
#ifndef MAIN
extern int no_map_backslashes;
int ls
#else
int no_map_backslashes = 0;
/*lint -e529 */
int dirp(s)
char *s;
{
	register int junk1,junk2;
	static fcb dir;
	/* handle all of the stupid special cases */
	if (0 == strcmp(&s[1],":") || /* root directory on a drive */
		0 == strcmp(&s[0],"/") ||
		0 == strcmp(&s[0],"\\") || /* root directory default drive */
		0 == strcmp(&s[1],":/") || /* root directory another drive */
		0 == strcmp(&s[1],":\\") ||
		0 == strcmp(&s[1],":.") ||
		0 == strcmp(&s[0],"."))
	{
		return 1;
	}
	if (0 == strcmp(s,".."))			/* parent of this directory */
	{
		int returnval;
		char *current,*getcwd();
		current = getcwd(NULL,64);		/* get current directory */
		/* everything except root has a parent */
		returnval = strlen(current) != 0;
		Free(current);
		return returnval;
	}
    /* set the disk transfer address */
    bdos(0x1A,&dir,0);
    /* do a search first for the directory path */
    return (bdos(0x4E,s,0x10) == 0 && bdos(0x4E,s,0) != 0);
}
int main
#endif
(argc,argv)
char *argv[];
{
	int noargs;
	int returnval;
	char *current;
	char namebuf[128];
	char *abspath(),*ap;
	/* 
	 * initialize statics
	 */
	mode = 0x10;
	lsverbose=0;column=4;recurse=0;
	lsquiet = 0;
	drivenum = 0;
	sortype = ALPHA;
	pf = printf;
	/*
	 * get current time
	 */
	year = (int)(((unsigned long)time((long *)NULL) >> 25) & 0x7F) + 80;
	/*
	 * set up a default search name
	 */
	if (noargs = (argc == 1))
		argc++;
	while(--argc)
	{
		if (noargs)
			current = "*.*";
		else
			current = *(++argv);	/* get current file name */
		if (*current == '-')
		{
			++current;	/* point past - */
			while (*current)
			{
				switch (*current++)
				{
				case 'l':
				case 'L':
					lsverbose = 1;
					if (column != 1)
						column = 2;
					if (lsquiet)
					{
						fprintf(stderr,"ls : lsverbose and quiet conflict\n");
						return -1;
					}
					break;
				case 'q':
				case 'Q':
					lsquiet = 1;
					if (lsverbose)
					{
						fprintf(stderr,"ls : quiet and lsverbose conflict\n");
						return -1;
					}
					break;
				case 'c':
				case 'C':
					column = 1;
					break;
				case 't':
				case 'T':
					sortype = RTIME;
					break;
				case 'a':
				case 'A':
					mode = 0x2 + 0x4 + 0x10;
					break;
				case 'R':
					recurse = 1;
					mode = 0x2 + 0x4 + 0x10;
					break;
				case 'r':
				/* toggle sort type */
					sortype ^= 1;
					break;
				default:
					break;
				}
			}
			/* if we're down to one argument after looking at all the
			   switches, we need to set noargs to true */
			if (noargs = (argc == 1))
				argc++;
			continue;
		}
		if (NULL == (ap = abspath(current)))
		{
			fprintf(stderr,"ls : bad path %s\n",current);
			return -1;
		}
		strcpy(namebuf,ap);
		drivenum = toupper(namebuf[0]) - 'A' + 1;
		if (!noargs && dirp(namebuf))
			if (strchr("\\/",namebuf[strlen(namebuf)-1]) != NULL)
				strcat(namebuf,"*.*");
			else
				strcat(namebuf,no_map_backslashes ? "\\*.*" : "/*.*");
		returnval = do_dir(namebuf);
	}
	return returnval;
}

jmp_buf do_dir_buf;
void	do_dir_sig()
{
	signal(SIGINT,SIG_IGN);	/* turn off ctrl-breaks */
	fprintf(stderr,"ls : interrupted\n");
	longjmp(do_dir_buf,-1);
}

int do_dir(current)
	char *current;
{
	file_desc *files = NULL;
	file_desc *curr_file,*getnext();
	unsigned int ftime,date;
	int i,j,col;
	long total = 0;
	char atts[6]; /* drwsa */

	/* do signal handling */
	if (-1 == setjmp(do_dir_buf))
	{
		if (files != NULL)
			Free(files);
		return -1;
	}
	signal(SIGINT,do_dir_sig);

	/* allocate file block */
	if (NULL == (files = Malloc((unsigned)sizeof(file_desc))))
	{
		fprintf(stderr,"Not enough memory to do directory\n");
		return -1;
	}
	/* look for match */
	i = 0;
	if (NULL == (curr_file = getfirst(current)))
	{
		fprintf(stderr,"ls : no files matching %s\n",current);
		(void)Free(files);
		return -1;	
	}
	files[i++] = *curr_file;
	/* i = 1 */
	/* get all matching */
	while ((curr_file = getnext()) && i < maxfiles)
	{
		(void)Free(files);	/* in case call to realloc fails */
		/* i = number of files - 1 */
		if (NULL == (files = realloc(files,sizeof(file_desc)*(i+1))))
		{
			fprintf(stderr,"Not enough memory to do directory\n");
			return -1;
		}
		files[i++] = *curr_file;	
	}
	if (i > 1)
		qsort(files,(unsigned)i,sizeof(file_desc),compare[sortype]);
	if (!lsquiet)
	{
		(*pf)("\n%s\n",current);
	}
	col = 1;
	for (j = 0; j < i; j++)
	{
		register char *c = files[j].file_name;
		if (*c == '.')
			continue;	/* filter out . and .. */
		while (*c)
		{
			*c = tolower(*c);
			c++;
		}
		if (lsverbose)
		{
			register char att = files[j].attribute;
			register unsigned int fyear;
			fyear = ((files[j].file_date >> 9) & 0x7F)+80; 
			atts[5] = 0;	/* terminate string */
			atts[0] = att & 0x10 ? 'd' : '-';
			atts[1] = att & 2 ? '-' : 'r';
			atts[2] = att & 1 ? '-' : 'w';
			atts[3] = att & 4 ? 's' : '-';
			atts[4] = att & 0x20 ? 'a' : '-';
			if (atts[0] == 'd')
			{
				register int k;
				(*pf)("%s %s/",atts,files[j].file_name);
				k = 12 - strlen(files[j].file_name)+7;
				while(k--)
					(*pf)(" ");
			}
			else
			{
				total += files[j].file_size;
				(*pf)("%s %-12s %-6ld ",atts,files[j].file_name,
						files[j].file_size);
			}
			ftime = files[j].file_time;
			date = files[j].file_date;
			if (year == fyear)
			{
				(*pf)("%02d/%02d %02d:%02d ",
					((date >> 5) & 0x0F),	/* month */
					date & 0x1F,		/* day	*/
					(ftime >> 11) & 0x1F,		/* hours */
					(ftime >> 5) & 0x3F);		/* minutes */
			}
			else
			{
				(*pf)("%02d/%02d       ",
					((date >> 5) & 0x0F),	/* month */
					fyear,					/* file year */
					(ftime >> 11) & 0x1F,		/* hours */
					(ftime >> 5) & 0x3F);		/* minutes */
			}
		}
		else
		{
			if (files[j].attribute & 0x10)
			{
				register int k;
				(*pf)("%s%c",
					files[j].file_name,(no_map_backslashes?'\\':'/'));
				k = 16 - strlen(files[j].file_name);
				while(--k)
					(*pf)(" ");
			}
			else
			{
				(*pf)("%-13s   ",files[j].file_name);
			}
		}
		if (col == column)
		{
			col = 1;
			(*pf)("\n");
		}
		else
			col++;
	}
	(*pf)("\n");
	if (lsverbose)
	{
		(*pf)("%ld bytes in %d files ",total,i);
		pr_Freespace();
	}
	if (recurse)
		for (j = 0; j < i; j++)
		{
			/* we've got a subdirectory */
			if (files[j].attribute & 0x10 && files[j].file_name[0] != '.')
			{
				char *path;
				char *dirname;
				int i;
				if (NULL == (dirname = Malloc(65)))
				{
					fprintf(stderr,"memory exhausted");
					goto freefiles;
				}
				for (i = 0; current[i] != '\0' && 
					NULL == strchr("*?",current[i]); i++)
					dirname[i] = current[i];
				dirname[i] = '\0';
				if ((path = strrchr(dirname,'/')) || 
					(path=strchr(dirname,'\\')))
					*(++path) = '\0';
				strcat(dirname,files[j].file_name);	/* get name */
				/* add wild cards */
				strcat(dirname,no_map_backslashes ? "\\*.*" : "/*.*");
				if (-1 == do_dir(dirname))
				{
					(void)Free(dirname);
				freefiles:
					(void)Free(files);
					return -1;
				}
				(void)Free(dirname);
			}
		}
	(void)Free(files);
	return 0;
}

int files_cmp(a,b)
	file_desc *a,*b;
{
	return strcmp(a->file_name,b->file_name);
}
int rfiles_cmp(a,b)
	file_desc *a,*b;
{
	int returnval = strcmp(a->file_name,b->file_name);
	return (returnval > 0 ? -1 : (returnval < 0 ? 1 : 0));
}
typedef unsigned long ulong;
int times_cmp(a,b)
	register file_desc *a,*b;
{
	long atime = (ulong)a->file_time | ((ulong)a->file_date << 16);
	long btime = (ulong)b->file_time | ((ulong)b->file_date << 16);
	if (atime < btime)
		return -1;
	else if (atime > btime)
		return 1;
	else return 0;
}
int rtimes_cmp(a,b)
	register file_desc *a,*b;
{
	long atime = (ulong)a->file_time | ((ulong)a->file_date << 16);
	long btime = (ulong)b->file_time | ((ulong)b->file_date << 16);
	if (atime > btime)
		return -1;
	else if (atime < btime)
		return 1;
	else return 0;
}

fcb tmp;

#ifdef MSC
static file_desc temp2;
#endif

static struct registers regs;

file_desc *getfirst(fname)
	char *fname;
{
	register int result;
	/* set the disk transfer address */
#ifdef MSC
	GETSEGS(&regs);
#endif
#ifdef AZTEC86
	extern int _dsval;
	regs.ds = _dsval;
#endif
	bdos(0x1A,&tmp,0);
	regs.ax = 0x4e00;
	regs.dx = (unsigned) fname;
	regs.cx = mode;
	/* make the find first call */
	(void)sysint21(&regs,&regs);
	if(2 == regs.ax || 18 == regs.ax)
		return NULL;
#ifdef AZTEC86
	return (file_desc *)&(tmp.attribute);
#endif
#ifdef MSC
	/* kludge to get around structure alignment problems */
	temp2.attribute = tmp.attribute;
	temp2.file_time = tmp.file_time;
	temp2.file_date = tmp.file_date;
	temp2.file_size = tmp.file_size;
	strcpy(temp2.file_name,tmp.file_name);
	return &temp2;
#endif
}

file_desc *getnext()
{
	register int result;
#ifdef AZTEC86
	extern int _dsval;
	regs.ds = _dsval;
#endif
#ifdef MSC
	GETSEGS(&regs);
#endif
	/* set the disk transfer address */
	bdos(0x1A,&tmp,0);
	regs.ax = 0x4f00;
	sysint21(&regs,&regs);
	/* make the find next call */
	if (18 == regs.ax)
		return NULL;
#ifdef AZTEC86
	return (file_desc *)&(tmp.attribute);
#endif
#ifdef MSC
	/* kludge to get around structure alignment problems */
	temp2.attribute = tmp.attribute;
	temp2.file_time = tmp.file_time;
	temp2.file_date = tmp.file_date;
	temp2.file_size = tmp.file_size;
	strcpy(temp2.file_name,tmp.file_name);
	return &temp2;
#endif
}

char *get_first_name(name)
	char *name;
{
	file_desc *x;
	if (NULL == (x = getfirst(name)))
		return NULL;
	return x->file_name;
}

char *get_next_name()
{
	file_desc *x;
	if (NULL == (x = getnext()))
		return NULL;
	return x->file_name;
}

/*	determine available space on default drive
*/
#ifndef MAIN
int du(argc,argv)
	char **argv;
{
	if (argc == 2)
	{
		char driveletter = toupper(argv[1][0]);
		if (driveletter < 'A' || driveletter > 'Z')
		{
			fprintf(stderr,"du : bad drive %s",argv[1]);
			return -1;
		}
		drivenum = toupper(argv[1][0]) - 'A' + 1;
	}
	else
		drivenum = 0;
	pr_Freespace();
	return 0;
}
#endif
pr_Freespace()
{
	/* register arguments for INT */
    struct   {int ax,bx,cx,dx,si,di,ds,es;}sysr;

#define 	    cls_avail  sysr.bx	    /* number of available clusters */
#define 	    cls_total  sysr.dx	    /* total number of clusters on volume */
#define 	    byt_sectr  sysr.cx	    /* bytes per sector */
#define 	    sec_clstr  sysr.ax	    /* sectors per cluster */

    unsigned long   byt_clstr,		    /* bytes per cluster */
		    vol_total,		    /* total size of volume, bytes */
		    fre_total;		    /* size of Free space, bytes */

    sysr.ax = 0x3600;		    /* get disk Free space function code */
    sysr.dx = drivenum;			    /* (DL) = drive id, 0=current, 1=A, */
	(void)sysint21(&sysr,&sysr);		    /* invoke DOS */
    if( sec_clstr!=0xffff ) 
	{
		byt_clstr = byt_sectr * sec_clstr;
		vol_total = cls_total * byt_clstr;
		fre_total = cls_avail * byt_clstr;
		(*pf)("%lu out of %lu bytes available on %c:\n", 
		fre_total,vol_total, 
		drivenum ? drivenum + 'A' - 1 : bdos(0x19,0,0)+'A');
	}
}
