# MAKEFILE for PC-SHELL
# MICROSOFT C VERSION
# macros to hash paths to commonly used programs
#
.SUFFIXES : .obj .c .asm
CC=msc
AS=masm
LN=LINK /CO
DEFSYMBOLS = -DRAINBOW -DMSC
CFLAGS=$(DEFSYMBOLS) /Zi
AFLAGS=
VERSION=3.5
# make rules
.c.obj:
	$(CC) $(CFLAGS) $* ,;
.asm.obj:
	masm /ML/B63/DMSC $*,;
# home for all commands
BINDIR=/bin/


# CFLAGS=+1
# small model library path
CLIB=-l/clibs/c

# source files for shell
shsrc1=main.c cp.c more.c cmds.c chmod.c y.c
shsrc2=ls.c fgrep.c mv.c rm.c dump2.c forkp.c
shsrc3=savestr.c _croot.c env.c cat.c ntharg.c abspath.c
shsrc4=mvcp.c dirfuncs.c misc.c doprompt.c expand.c
shsrc=$(shsrc1) $(shsrc2) $(shsrc3) $(shsrc4)

shell.lnt : makefile
	echo -u -i/include/ -eau -e525 -e718 -e720 +fsa +fvo >$@
	echo +fzu /include/stdlib >>$@
	echo $(shsrc1) >>$@
	echo $(shsrc2) >>$@
	echo $(shsrc3) >>$@
	echo $(shsrc4) >>$@

# assembly source for shell
shasm=fexec.asm ctlbrk.asm crit.asm

editall :
	e $(shsrc1)
	e $(shsrc2)
	e $(shsrc3)
	e $(shsrc4)
	e $(shasm)

#
# check in everything with version control
VCSFLAGS=-N -M
checkin :
	put $(VCSFLAGS) vcs(*.c) vcs(*.asm) vcs(*.h) vcs(makefile)

# object files for shell
shobj1=main.obj cp.obj more.obj cmds.obj chmod.obj y.obj
shobj2=ls.obj fgrep.obj mv.obj rm.obj dump2.obj forkp.obj
shobj3=savestr.obj _croot.obj env.obj cat.obj ntharg.obj abspath.obj
shobj4=mvcp.obj dirfuncs.obj misc.obj doprompt.obj expand.obj
asmobj=crit.obj ctlbrk.obj
shobj = $(shobj1) $(shobj2) $(shobj3) $(shobj4) $(asmobj)
#
# main.obj depends on shell.h so that changes in global variables
# contained there will always force a recompile
main.obj : shell.h


# source files for the small shell
smsrc1=main.c cmds.c chmod.c y.c
smsrc2=rm.c forkp.c cmdlist.c
smsrc3=savestr.c _croot.c env.c cat.c ntharg.c abspath.c
smsrc4=dirfuncs.c misc.c doprompt.c expand.c
smsrc=$(shsrc1) $(shsrc2) $(shsrc3) $(shsrc4)

# object files for small shell
smobj1=main.obj cmds.obj chmod.obj y.obj fexec.obj
smobj2=rm.obj forkp.obj
smobj3=savestr.obj _croot.obj env.obj cat.obj ntharg.obj crit.obj abspath.obj
smobj4=ctlbrk.obj dirfuncs.obj misc.obj doprompt.obj expand.obj
smobj = $(smobj1) $(smobj2) $(smobj3) $(smobj4)

small.lnk : makefile
	echo $(smobj1) > $@
	echo $(smobj2) >> $@
	echo $(smobj3) >> $@
	echo $(smobj4) >> $@
	echo $(LIB) >> $@

#
# dependencies for debugging malloc and free
env.obj main.obj _croot.obj mvcp.obj ls.obj savestr.obj expand.obj : mem.h 

#
# linting using pc-lint
lint : shell.lnt
	lint shell.lnt >lint.objut

#
# make main.c from state machine source
# main.c : main.s
#	statecom main

#
# make pc-shell - check currency of all percursors, link and then move to
# bin directory
#
pc-shell.exe : $(shobj) shell.lnk
	chmod +w $@
	LINK $(shobj),$@ /CO ,$*/MAP;
	chmod -w $@

#
# small cmdlist
smcmd.obj : cmdlist.c
	cc $(CFLAGS) -dSMALL cmdlist -o $@

#
# small pc-shell
$(BINDIR)small.com : $(smobj) smcmd.obj stksiz.obj small.lnk
	chmod +w $@
	$(LN) -t -o $@ smcmd.obj stksiz.obj -f small.lnk
	chmod -w $@

shell.doc : shell.nr
	nro -mm shell.nr >shell.doc


#
# big enchilada archive of all source
arc :
	pkarc u /release/shell.arc *.c *.h *.nr *.asm $(BINDIR)pc-shell.com

#
# make a release package
# the obj directory is a child of the current working directory
release :
	@make production
	@rm *.bak *.sym *.map *.dbg lint.objut shell.doc >NUL
	@make arc

ns= pc-shell.com shell.doc

nosource :
	@make production
	@make shell.doc
	pkarc u /arc/shellns $(ns)

production : 
	make $(BINDIR)pc-shell.com
	@-cp $(BINDIR)pc-shell.com .

#
# make entries to produce stand-alone versions of shell builtins
#

mainrm.obj : rm.c
	$(CC) -dRAINBOW -dMAIN $< -o $@

$(BINDIR)rm.com : croot.obj mainrm.obj
	chmod +w $@
	$(LN) -o $@ $< -l/clibs/c

mainchmod.obj : chmod.c
	$(CC) -dMAIN $< -o $@

$(BINDIR)chmod.com : croot.obj mainchmod.obj
	chmod +w $@
	$(LN) -o $@ $< -l/clibs/c
	chmod -w $@

mainmore.obj : more.c
	$(CC) -dMAIN -dRAINBOW $< -o $@

$(BINDIR)more.com : mainmore.obj
	chmod +w $@
	$(LN) -o $@ $< -l/clibs/c
	chmod -w $@

mainmv.obj : mv.c
	$(CC) -dMAIN $< -o $@

$(BINDIR)mv.com : croot.obj savestr.obj mvcp.obj mainmv.obj abspath.obj
	chmod +w $@
	$(LN) -g -t -o $@ $< -l/clibs/c
	chmod -w $@

maindump.obj : dump2.c
	$(CC) -dMAIN $< -o $@

$(BINDIR)dump.com : croot.obj maindump.obj
	chmod +w $@
	$(LN) -o $@ $< -l/clibs/c
	chmod -w $@


maincp.obj : cp.c
	$(CC) -dMAIN $< -o $@

$(BINDIR)cp.com : croot.obj savestr.obj mvcp.obj maincp.obj abspath.obj
	chmod +w $@
	$(LN) -o $@ $< -l/clibs/c
	chmod -w $@

mainls.obj : ls.c
	$(CC) -dMAIN -dRAINBOW $< -o $@

$(BINDIR)ls.com : mainls.obj abspath.obj
	chmod +w $@
	$(LN) -t -o $@ $< -l/clibs/c
	chmod -w $@

#
# all the standalones
CMDS = /bin/chmod.com /bin/more.com /bin/mv.com /bin/dump.com \
/bin/cp.com /bin/ls.com /bin/rm.com

ACMDS = chmod.com more.com mv.com dump.com \
cp.com ls.com rm.com

/release/cmds.arc : $(CMDS) .PRECIOUS
	cd \bin
	pkarc u $@ $(ACMDS)
	cd \shell

clean :
	rm *.doc *.bak *.sym *.dbg lint.objut

